//
//  ESSIPEService.h
//  SIPEAdiumPlugin
//
//  Created by Matt Meissner on 10/30/09.
//  Modified by Michael Lamb on 2/27/13
//  Copyright 2013 Michael Lamb/Harris Kauffman. All rights reserved.
//

//#import <Adium/AIService.h>
#import <AdiumLibpurple/PurpleService.h>

@interface ESSIPEService : PurpleService {
    
}

@end
