// Generated by ResEdit 1.5.4
// Copyright (C) 2006-2010
// http://www.resedit.net

#include <windows.h>
#include <commctrl.h>
#include <richedit.h>
#include "miranda-resource.h"




//
// Dialog resources
//
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
IDD_ACCMGRUI DIALOGEX 0, 0, 186, 134
STYLE DS_SHELLFONT | WS_VISIBLE | WS_CHILDWINDOW
EXSTYLE WS_EX_CONTROLPARENT
FONT 8, "MS Shell Dlg", 400, 0, 1
{
    LTEXT           "Username:", IDC_STATIC, 5, 5, 45, 12, SS_LEFT
    EDITTEXT        IDC_HANDLE, 50, 5, 116, 12, ES_AUTOHSCROLL
    AUTORADIOBUTTON "Use windows credentials", IDC_SSO, 5, 20, 94, 8
    AUTORADIOBUTTON "Manually specify parameters", IDC_MSO, 5, 35, 105, 8
    LTEXT           "Login:", IDC_STATIC, 20, 50, 45, 12, SS_LEFT
    EDITTEXT        IDC_LOGIN, 65, 50, 116, 12, ES_AUTOHSCROLL
    LTEXT           "Password:", IDC_STATIC, 20, 65, 45, 12, SS_LEFT
    EDITTEXT        IDC_PASSWORD, 65, 65, 116, 12, ES_AUTOHSCROLL | ES_PASSWORD
}



LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
IDD_SEARCHUI DIALOGEX 0, 0, 109, 148
STYLE DS_SHELLFONT | WS_VISIBLE | WS_CHILDWINDOW
EXSTYLE WS_EX_CONTROLPARENT
FONT 8, "Ms Shell Dlg"
{
    LTEXT           "First Name", IDC_STATIC, 3, 5, 44, 12, SS_LEFT
    EDITTEXT        IDC_SEARCH_FN, 3, 15, 104, 12, ES_AUTOHSCROLL
    LTEXT           "Last Name", IDC_STATIC, 3, 30, 44, 12, SS_LEFT
    EDITTEXT        IDC_SEARCH_LN, 3, 40, 104, 12, ES_AUTOHSCROLL
    LTEXT           "Company", IDC_STATIC, 3, 55, 44, 12, SS_LEFT
    EDITTEXT        IDC_SEARCH_COMPANY, 3, 65, 104, 12, ES_AUTOHSCROLL
    LTEXT           "Country", IDC_STATIC, 3, 80, 44, 12, SS_LEFT
    EDITTEXT        IDC_SEARCH_COUNTRY, 3, 90, 104, 12, ES_AUTOHSCROLL
}



LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
IDD_OPT_SIPSIMPLE DIALOG 0, 0, 255, 230
STYLE DS_SHELLFONT | WS_VISIBLE | WS_CHILDWINDOW
EXSTYLE WS_EX_CONTROLPARENT
FONT 8, "Ms Shell Dlg"
{
    GROUPBOX        "Account options", IDC_STATIC, 5, 5, 245, 70
    AUTOCHECKBOX    "Use windows credentials for authentication", IDC_USESSO, 15, 20, 155, 12
    LTEXT           "Authentication scheme", IDC_STATIC, 20, 35, 75, 12, SS_LEFT
    COMBOBOX        IDC_AUTHTYPE, 100, 35, 75, 30, CBS_DROPDOWNLIST | CBS_HASSTRINGS
    LTEXT           "Connection type", IDC_STATIC, 20, 55, 75, 12, SS_LEFT
    COMBOBOX        IDC_CONNTYPE, 100, 55, 75, 30, CBS_DROPDOWNLIST | CBS_HASSTRINGS
    GROUPBOX        "User information", IDC_STATIC, 5, 80, 245, 65
    LTEXT           "Username:", IDC_STATIC, 15, 95, 55, 12, SS_LEFT
    EDITTEXT        IDC_HANDLE, 75, 95, 135, 12, ES_AUTOHSCROLL
    LTEXT           "Login:", IDC_STATIC, 15, 110, 55, 12, SS_LEFT
    EDITTEXT        IDC_LOGIN, 75, 110, 135, 12, ES_AUTOHSCROLL
    LTEXT           "Password:", IDC_STATIC, 15, 125, 55, 12, SS_LEFT
    EDITTEXT        IDC_PASSWORD, 75, 125, 135, 12, ES_AUTOHSCROLL | ES_PASSWORD
    GROUPBOX        "Public IP (Protocol global)", IDC_STATIC, 5, 150, 245, 75
    AUTORADIOBUTTON "Fetch local IP", IDC_IPLOCAL, 15, 165, 55, 12
    LTEXT           "", IDC_IPLOCALFOUND, 90, 165, 55, 12, SS_LEFT
    AUTORADIOBUTTON "Set manually", IDC_IPMANUAL, 15, 180, 55, 12
    EDITTEXT        IDC_PUBLICIP, 90, 180, 135, 12, ES_AUTOHSCROLL
    AUTORADIOBUTTON "Run program", IDC_IPPROG, 15, 195, 55, 12
    EDITTEXT        IDC_IPPROGEXE, 90, 195, 135, 12, ES_AUTOHSCROLL
}



LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
IDD_OPT_SIPSIMPLE_ABOUT DIALOG 0, 0, 255, 230
STYLE DS_SHELLFONT | WS_VISIBLE | WS_CHILDWINDOW
EXSTYLE WS_EX_CONTROLPARENT
FONT 8, "Ms Shell Dlg"
{
    CONTROL         "Rich Edit", IDC_ABOUTSIPE, RICHEDIT_CLASS, WS_TABSTOP | WS_BORDER | ES_NUMBER | ES_MULTILINE | ES_READONLY | ES_AUTOVSCROLL | WS_VSCROLL, 5, 8, 240, 216
}
